# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hmac
import time
import base64
import hashlib
from io import FileIO as file

from libcloud.utils.py3 import b, next, httplib, urlparse, urlquote, urlencode, urlunquote
from libcloud.common.base import XmlResponse, ConnectionUserAndKey
from libcloud.utils.files import read_in_chunks
from libcloud.common.types import LibcloudError
from libcloud.storage.base import CHUNK_SIZE, Object, Container, StorageDriver
from libcloud.storage.types import (
    ObjectDoesNotExistError,
    ContainerIsNotEmptyError,
    ContainerDoesNotExistError,
    ContainerAlreadyExistsError,
)


def collapse(s):
    return " ".join([x for x in s.split(" ") if x])


class AtmosError(LibcloudError):
    def __init__(self, code, message, driver=None):
        super().__init__(value=message, driver=driver)
        self.code = code


class AtmosResponse(XmlResponse):
    def success(self):
        return self.status in (
            httplib.OK,
            httplib.CREATED,
            httplib.NO_CONTENT,
            httplib.PARTIAL_CONTENT,
        )

    def parse_error(self):
        tree = self.parse_body()

        if tree is None:
            return None

        code = int(tree.find("Code").text)
        message = tree.find("Message").text
        raise AtmosError(code=code, message=message, driver=self.connection.driver)


class AtmosConnection(ConnectionUserAndKey):
    responseCls = AtmosResponse

    def add_default_headers(self, headers):
        headers["x-emc-uid"] = self.user_id
        headers["Date"] = time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())
        headers["x-emc-date"] = headers["Date"]

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/octet-stream"

        if "Accept" not in headers:
            headers["Accept"] = "*/*"

        return headers

    def pre_connect_hook(self, params, headers):
        headers["x-emc-signature"] = self._calculate_signature(params, headers)

        return params, headers

    def _calculate_signature(self, params, headers):
        pathstring = urlunquote(self.action)
        driver_path = self.driver.path  # pylint: disable=no-member

        if pathstring.startswith(driver_path):
            pathstring = pathstring[len(driver_path) :]

        if params:
            if type(params) is dict:
                params = list(params.items())
            pathstring += "?" + urlencode(params)
        pathstring = pathstring.lower()

        xhdrs = [(k, v) for k, v in list(headers.items()) if k.startswith("x-emc-")]
        xhdrs.sort(key=lambda x: x[0])

        signature = [
            self.method,
            headers.get("Content-Type", ""),
            headers.get("Range", ""),
            headers.get("Date", ""),
            pathstring,
        ]
        signature.extend([k + ":" + collapse(v) for k, v in xhdrs])
        signature = "\n".join(signature)
        key = base64.b64decode(self.key)
        signature = hmac.new(b(key), b(signature), hashlib.sha1).digest()

        return base64.b64encode(b(signature)).decode("utf-8")


class AtmosDriver(StorageDriver):
    connectionCls = AtmosConnection

    host = None  # type: str
    path = None  # type: str
    api_name = "atmos"
    supports_chunked_encoding = True
    website = "http://atmosonline.com/"
    name = "atmos"

    DEFAULT_CDN_TTL = 60 * 60 * 24 * 7  # 1 week

    def __init__(self, key, secret=None, secure=True, host=None, port=None):
        host = host or self.host
        super().__init__(key, secret, secure, host, port)

    def iterate_containers(self):
        result = self.connection.request(self._namespace_path(""))
        entries = self._list_objects(result.object, object_type="directory")

        for entry in entries:
            extra = {"object_id": entry["id"]}
            yield Container(entry["name"], extra, self)

    def get_container(self, container_name):
        path = self._namespace_path(container_name) + "/?metadata/system"
        try:
            result = self.connection.request(path)
        except AtmosError as e:
            if e.code != 1003:
                raise
            raise ContainerDoesNotExistError(e, self, container_name)
        meta = self._emc_meta(result)
        extra = {"object_id": meta["objectid"]}

        return Container(container_name, extra, self)

    def create_container(self, container_name):
        path = self._namespace_path(container_name) + "/"
        try:
            self.connection.request(path, method="POST")
        except AtmosError as e:
            if e.code != 1016:
                raise
            raise ContainerAlreadyExistsError(e, self, container_name)

        return self.get_container(container_name)

    def delete_container(self, container):
        try:
            self.connection.request(self._namespace_path(container.name) + "/", method="DELETE")
        except AtmosError as e:
            if e.code == 1003:
                raise ContainerDoesNotExistError(e, self, container.name)
            elif e.code == 1023:
                raise ContainerIsNotEmptyError(e, self, container.name)

        return True

    def get_object(self, container_name, object_name):
        container = self.get_container(container_name)
        object_name_cleaned = self._clean_object_name(object_name)
        path = self._namespace_path(container_name) + "/" + object_name_cleaned

        try:
            result = self.connection.request(path + "?metadata/system")
            system_meta = self._emc_meta(result)

            result = self.connection.request(path + "?metadata/user")
            user_meta = self._emc_meta(result)
        except AtmosError as e:
            if e.code != 1003:
                raise
            raise ObjectDoesNotExistError(e, self, object_name)

        last_modified = time.strptime(system_meta["mtime"], "%Y-%m-%dT%H:%M:%SZ")
        last_modified = time.strftime("%a, %d %b %Y %H:%M:%S GMT", last_modified)
        extra = {"object_id": system_meta["objectid"], "last_modified": last_modified}
        data_hash = user_meta.pop("md5", "")

        return Object(
            object_name,
            int(system_meta["size"]),
            data_hash,
            extra,
            user_meta,
            container,
            self,
        )

    def upload_object(
        self,
        file_path,
        container,
        object_name,
        extra=None,
        verify_hash=True,
        headers=None,
    ):
        method = "PUT"

        extra = extra or {}
        object_name_cleaned = self._clean_object_name(object_name)
        request_path = self._namespace_path(container.name) + "/" + object_name_cleaned
        content_type = extra.get("content_type", None)

        try:
            self.connection.request(request_path + "?metadata/system")
        except AtmosError as e:
            if e.code != 1003:
                raise
            method = "POST"

        result_dict = self._upload_object(
            object_name=object_name,
            content_type=content_type,
            request_path=request_path,
            request_method=method,
            headers={},
            file_path=file_path,
        )

        bytes_transferred = result_dict["bytes_transferred"]

        if extra is None:
            meta_data = {}
        else:
            meta_data = extra.get("meta_data", {})
        meta_data["md5"] = result_dict["data_hash"]
        user_meta = ", ".join([k + "=" + str(v) for k, v in list(meta_data.items())])
        self.connection.request(
            request_path + "?metadata/user",
            method="POST",
            headers={"x-emc-meta": user_meta},
        )
        result = self.connection.request(request_path + "?metadata/system")
        meta = self._emc_meta(result)
        del meta_data["md5"]
        extra = {
            "object_id": meta["objectid"],
            "meta_data": meta_data,
        }

        return Object(
            object_name,
            bytes_transferred,
            result_dict["data_hash"],
            extra,
            meta_data,
            container,
            self,
        )

    def upload_object_via_stream(self, iterator, container, object_name, extra=None, headers=None):
        if isinstance(iterator, file):
            iterator = iter(iterator)

        extra_headers = headers or {}
        data_hash = hashlib.md5()  # nosec
        generator = read_in_chunks(iterator, CHUNK_SIZE, True)
        bytes_transferred = 0
        try:
            chunk = next(generator)
        except StopIteration:
            chunk = ""

        path = self._namespace_path(container.name + "/" + object_name)
        method = "PUT"

        if extra is not None:
            content_type = extra.get("content_type", None)
        else:
            content_type = None

        content_type = self._determine_content_type(content_type, object_name)

        try:
            self.connection.request(path + "?metadata/system")
        except AtmosError as e:
            if e.code != 1003:
                raise
            method = "POST"

        while True:
            end = bytes_transferred + len(chunk) - 1
            data_hash.update(b(chunk))
            headers = dict(extra_headers)

            headers.update(
                {
                    "x-emc-meta": "md5=" + data_hash.hexdigest(),  # nosec
                    "Content-Type": content_type,
                }
            )

            if len(chunk) > 0 and bytes_transferred > 0:
                headers["Range"] = "Bytes=%d-%d" % (bytes_transferred, end)
                method = "PUT"

            result = self.connection.request(path, method=method, data=chunk, headers=headers)
            bytes_transferred += len(chunk)

            try:
                chunk = next(generator)
            except StopIteration:
                break

            if len(chunk) == 0:
                break

        data_hash = data_hash.hexdigest()  # nosec

        if extra is None:
            meta_data = {}
        else:
            meta_data = extra.get("meta_data", {})
        meta_data["md5"] = data_hash
        user_meta = ", ".join([k + "=" + str(v) for k, v in list(meta_data.items())])
        self.connection.request(
            path + "?metadata/user", method="POST", headers={"x-emc-meta": user_meta}
        )

        result = self.connection.request(path + "?metadata/system")

        meta = self._emc_meta(result)
        extra = {
            "object_id": meta["objectid"],
            "meta_data": meta_data,
        }

        return Object(object_name, bytes_transferred, data_hash, extra, meta_data, container, self)

    def download_object(
        self, obj, destination_path, overwrite_existing=False, delete_on_failure=True
    ):
        path = self._namespace_path(obj.container.name + "/" + obj.name)
        response = self.connection.request(path, method="GET", raw=True)

        return self._get_object(
            obj=obj,
            callback=self._save_object,
            response=response,
            callback_kwargs={
                "obj": obj,
                "response": response.response,
                "destination_path": destination_path,
                "overwrite_existing": overwrite_existing,
                "delete_on_failure": delete_on_failure,
            },
            success_status_code=httplib.OK,
        )

    def download_object_as_stream(self, obj, chunk_size=None):
        path = self._namespace_path(obj.container.name + "/" + obj.name)
        response = self.connection.request(path, method="GET", raw=True)

        return self._get_object(
            obj=obj,
            callback=read_in_chunks,
            response=response,
            callback_kwargs={"iterator": response.response, "chunk_size": chunk_size},
            success_status_code=httplib.OK,
        )

    def delete_object(self, obj):
        path = self._namespace_path(obj.container.name) + "/" + self._clean_object_name(obj.name)
        try:
            self.connection.request(path, method="DELETE")
        except AtmosError as e:
            if e.code != 1003:
                raise
            raise ObjectDoesNotExistError(e, self, obj.name)

        return True

    def enable_object_cdn(self, obj):
        return True

    def get_object_cdn_url(self, obj, expiry=None, use_object=False):
        """
        Return an object CDN URL.

        :param obj: Object instance
        :type  obj: :class:`Object`

        :param expiry: Expiry
        :type expiry: ``str``

        :param use_object: Use object
        :type use_object: ``bool``

        :rtype: ``str``
        """

        if use_object:
            path = "/rest/objects" + obj.meta_data["object_id"]
        else:
            path = "/rest/namespace/" + obj.container.name + "/" + obj.name

        if self.secure:
            protocol = "https"
        else:
            protocol = "http"

        expiry = str(expiry or int(time.time()) + self.DEFAULT_CDN_TTL)
        params = [
            ("uid", self.key),
            ("expires", expiry),
        ]
        params.append(("signature", self._cdn_signature(path, params, expiry)))

        params = urlencode(params)
        path = self.path + path

        return urlparse.urlunparse((protocol, self.host, path, "", params, ""))

    def _cdn_signature(self, path, params, expiry):
        key = base64.b64decode(self.secret)
        signature = "\n".join(["GET", path.lower(), self.key, expiry])
        signature = hmac.new(key, signature, hashlib.sha1).digest()

        return base64.b64encode(signature)

    def _list_objects(self, tree, object_type=None):
        listing = tree.find(self._emc_tag("DirectoryList"))
        entries = []

        for entry in listing.findall(self._emc_tag("DirectoryEntry")):
            file_type = entry.find(self._emc_tag("FileType")).text

            if object_type is not None and object_type != file_type:
                continue
            entries.append(
                {
                    "id": entry.find(self._emc_tag("ObjectID")).text,
                    "type": file_type,
                    "name": entry.find(self._emc_tag("Filename")).text,
                }
            )

        return entries

    def _clean_object_name(self, name):
        return urlquote(name.encode("ascii"))

    def _namespace_path(self, path):
        return self.path + "/rest/namespace/" + urlquote(path.encode("ascii"))

    def _object_path(self, object_id):
        return self.path + "/rest/objects/" + object_id.encode("ascii")

    @staticmethod
    def _emc_tag(tag):
        return "{http://www.emc.com/cos/}" + tag

    def _emc_meta(self, response):
        meta = response.headers.get("x-emc-meta", "")

        if len(meta) == 0:
            return {}
        meta = meta.split(", ")

        return dict([x.split("=", 1) for x in meta])

    def _entries_to_objects(self, container, entries):
        for entry in entries:
            metadata = {"object_id": entry["id"]}
            yield Object(entry["name"], 0, "", {}, metadata, container, self)

    def iterate_container_objects(self, container, prefix=None, ex_prefix=None):
        """
        Return a generator of objects for the given container.

        :param container: Container instance
        :type container: :class:`Container`

        :param prefix: Filter objects starting with a prefix.
                       Filtering is performed client-side.
        :type  prefix: ``str``

        :param ex_prefix: (Deprecated.) Filter objects starting with a prefix.
                          Filtering is performed client-side.
        :type  ex_prefix: ``str``

        :return: A generator of Object instances.
        :rtype: ``generator`` of :class:`Object`
        """
        prefix = self._normalize_prefix_argument(prefix, ex_prefix)

        headers = {"x-emc-include-meta": "1"}
        path = self._namespace_path(container.name) + "/"
        result = self.connection.request(path, headers=headers)
        entries = self._list_objects(result.object, object_type="regular")
        objects = self._entries_to_objects(container, entries)

        return self._filter_listed_container_objects(objects, prefix)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Vultr DNS Driver
"""

import json
from typing import Any, Dict, List, Optional

from libcloud.dns.base import Zone, Record, DNSDriver
from libcloud.dns.types import (
    Provider,
    RecordType,
    ZoneDoesNotExistError,
    ZoneAlreadyExistsError,
    RecordDoesNotExistError,
    RecordAlreadyExistsError,
)
from libcloud.utils.py3 import urlencode
from libcloud.common.vultr import (
    DEFAULT_API_VERSION,
    VultrResponse,
    VultrConnection,
    VultrResponseV2,
    VultrConnectionV2,
)

__all__ = [
    "ZoneRequiredException",
    "VultrDNSResponse",
    "VultrDNSConnection",
    "VultrDNSDriver",
]


class ZoneRequiredException(Exception):
    pass


class VultrDNSResponse(VultrResponse):
    pass


class VultrDNSConnection(VultrConnection):
    responseCls = VultrDNSResponse


class VultrDNSResponseV2(VultrResponseV2):
    pass


class VultrDNSConnectionV2(VultrConnectionV2):
    responseCls = VultrDNSResponseV2


class VultrDNSDriver(DNSDriver):
    type = Provider.VULTR
    name = "Vultr DNS"
    website = "https://www.vultr.com"

    def __new__(
        cls,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        api_version=DEFAULT_API_VERSION,
        region=None,
        **kwargs,
    ):
        if cls is VultrDNSDriver:
            if api_version == "1":
                cls = VultrDNSDriverV1
            elif api_version == "2":
                cls = VultrDNSDriverV2
            else:
                raise NotImplementedError(
                    "No Vultr driver found for API version: %s" % (api_version)
                )

        return super().__new__(cls)


class VultrDNSDriverV1(VultrDNSDriver):
    connectionCls = VultrDNSConnection

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.TXT: "TXT",
        RecordType.CNAME: "CNAME",
        RecordType.MX: "MX",
        RecordType.NS: "NS",
        RecordType.SRV: "SRV",
    }

    def list_zones(self):
        """
        Return a list of records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        action = "/v1/dns/list"
        params = {"api_key": self.key}
        response = self.connection.request(action=action, params=params)
        zones = self._to_zones(response.objects[0])

        return zones

    def list_records(self, zone):
        """
        Returns a list of records for the provided zone.

        :param zone: zone to list records for
        :type zone: `Zone`

        :rtype: list of :class: `Record`
        """

        if not isinstance(zone, Zone):
            raise ZoneRequiredException("zone should be of type Zone")

        zones = self.list_zones()

        if not self.ex_zone_exists(zone.domain, zones):
            raise ZoneDoesNotExistError(value="", driver=self, zone_id=zone.domain)

        action = "/v1/dns/records"
        params = {"domain": zone.domain}
        response = self.connection.request(action=action, params=params)
        records = self._to_records(response.objects[0], zone=zone)

        return records

    def get_zone(self, zone_id):
        """
        Returns a `Zone` instance.

        :param zone_id: name of the zone user wants to get.
        :type zone_id: ``str``

        :rtype: :class:`Zone`
        """
        ret_zone = None

        action = "/v1/dns/list"
        params = {"api_key": self.key}
        response = self.connection.request(action=action, params=params)
        zones = self._to_zones(response.objects[0])

        if not self.ex_zone_exists(zone_id, zones):
            raise ZoneDoesNotExistError(value=None, zone_id=zone_id, driver=self)

        for zone in zones:
            if zone_id == zone.domain:
                ret_zone = zone

        return ret_zone

    def get_record(self, zone_id, record_id):
        """
        Returns a Record instance.

        :param zone_id: name of the required zone
        :type zone_id: ``str``

        :param record_id: ID of the required record
        :type record_id: ``str``

        :rtype: :class: `Record`
        """
        ret_record = None
        zone = self.get_zone(zone_id=zone_id)
        records = self.list_records(zone=zone)

        if not self.ex_record_exists(record_id, records):
            raise RecordDoesNotExistError(value="", driver=self, record_id=record_id)

        for record in records:
            if record_id == record.id:
                ret_record = record

        return ret_record

    def create_zone(self, domain, type="master", ttl=None, extra=None):
        """
        Returns a `Zone` object.

        :param domain: Zone domain name, (e.g. example.com).
        :type domain: ``str``

        :param type: Zone type (master / slave).
        :type  type: ``str``

        :param ttl: TTL for new records. (optional)
        :type  ttl: ``int``

        :param extra: (optional) Extra attributes (driver specific).
                      (e.g. {'serverip':'127.0.0.1'})
        """
        extra = extra or {}

        if extra and extra.get("serverip"):
            serverip = extra["serverip"]
        else:
            raise ValueError("Missing servertip key in extra")

        params = {"api_key": self.key}
        data = urlencode({"domain": domain, "serverip": serverip})
        action = "/v1/dns/create_domain"
        zones = self.list_zones()

        if self.ex_zone_exists(domain, zones):
            raise ZoneAlreadyExistsError(value="", driver=self, zone_id=domain)

        self.connection.request(params=params, action=action, data=data, method="POST")
        zone = Zone(id=domain, domain=domain, type=type, ttl=ttl, driver=self, extra=extra)

        return zone

    def create_record(self, name, zone, type, data, extra=None):
        """
        Create a new record.

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param zone: Zone where the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :param extra: Extra attributes (driver specific). (optional)
        :type extra: ``dict``

        :rtype: :class:`Record`
        """
        extra = extra or {}

        ret_record = None
        old_records_list = self.list_records(zone=zone)
        # check if record already exists
        # if exists raise RecordAlreadyExistsError

        for record in old_records_list:
            if record.name == name and record.data == data:
                raise RecordAlreadyExistsError(value="", driver=self, record_id=record.id)

        MX = self.RECORD_TYPE_MAP.get("MX")
        SRV = self.RECORD_TYPE_MAP.get("SRV")

        if extra and extra.get("priority"):
            priority = int(extra["priority"])
        else:
            priority = None

        post_data = {
            "domain": zone.domain,
            "name": name,
            "type": self.RECORD_TYPE_MAP.get(type),
            "data": data,
        }

        if type == MX or type == SRV:
            if priority is None:
                raise ValueError("Missing priority argument for MX record type")
            post_data["priority"] = priority

        encoded_data = urlencode(post_data)
        params = {"api_key": self.key}
        action = "/v1/dns/create_record"

        self.connection.request(action=action, params=params, data=encoded_data, method="POST")
        updated_zone_records = zone.list_records()

        for record in updated_zone_records:
            if record.name == name and record.data == data:
                ret_record = record

        return ret_record

    def delete_zone(self, zone):
        """
        Delete a zone.

        Note: This will delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        action = "/v1/dns/delete_domain"
        params = {"api_key": self.key}
        data = urlencode({"domain": zone.domain})
        zones = self.list_zones()

        if not self.ex_zone_exists(zone.domain, zones):
            raise ZoneDoesNotExistError(value="", driver=self, zone_id=zone.domain)

        response = self.connection.request(params=params, action=action, data=data, method="POST")

        return response.status == 200

    def delete_record(self, record):
        """
        Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``
        """
        action = "/v1/dns/delete_record"
        params = {"api_key": self.key}
        data = urlencode({"RECORDID": record.id, "domain": record.zone.domain})

        zone_records = self.list_records(record.zone)

        if not self.ex_record_exists(record.id, zone_records):
            raise RecordDoesNotExistError(value="", driver=self, record_id=record.id)

        response = self.connection.request(action=action, params=params, data=data, method="POST")

        return response.status == 200

    def ex_zone_exists(self, zone_id, zones_list):
        """
        Function to check if a `Zone` object exists.

        :param zone_id: Name of the `Zone` object.
        :type zone_id: ``str``

        :param zones_list: A list containing `Zone` objects
        :type zones_list: ``list``

        :rtype: Returns `True` or `False`
        """

        zone_ids = []

        for zone in zones_list:
            zone_ids.append(zone.domain)

        return zone_id in zone_ids

    def ex_record_exists(self, record_id, records_list):
        """
        :param record_id: Name of the `Record` object.
        :type record_id: ``str``

        :param records_list: A list containing `Record` objects
        :type records_list: ``list``

        :rtype: ``bool``
        """
        record_ids = []

        for record in records_list:
            record_ids.append(record.id)

        return record_id in record_ids

    def _to_zone(self, item):
        """
        Build an object `Zone` from the item dictionary

        :param item: item to build the zone from
        :type item: `dictionary`

        :rtype: :instance: `Zone`
        """
        type = "master"
        extra = {"date_created": item["date_created"]}

        zone = Zone(
            id=item["domain"],
            domain=item["domain"],
            driver=self,
            type=type,
            ttl=None,
            extra=extra,
        )

        return zone

    def _to_zones(self, items):
        """
        Returns a list of `Zone` objects.

        :param: items: a list that contains dictionary objects to be passed
        to the _to_zone function.
        :type items: ``list``
        """
        zones = []

        for item in items:
            zones.append(self._to_zone(item))

        return zones

    def _to_record(self, item, zone):
        extra = {}

        if item.get("priority"):
            extra["priority"] = item["priority"]

        type = self._string_to_record_type(item["type"])
        record = Record(
            id=item["RECORDID"],
            name=item["name"],
            type=type,
            data=item["data"],
            zone=zone,
            driver=self,
            extra=extra,
        )

        return record

    def _to_records(self, items, zone):
        records = []

        for item in items:
            records.append(self._to_record(item, zone=zone))

        return records


class VultrDNSDriverV2(VultrDNSDriver):
    connectionCls = VultrDNSConnectionV2

    RECORD_TYPE_MAP = {
        RecordType.A: "A",
        RecordType.AAAA: "AAAA",
        RecordType.CNAME: "CNAME",
        RecordType.NS: "NS",
        RecordType.MX: "MX",
        RecordType.SRV: "SRV",
        RecordType.TXT: "TXT",
        RecordType.CAA: "CAA",
        RecordType.SSHFP: "SSHFP",
    }

    def list_zones(self) -> List[Zone]:
        """Return a list of zones.

        :return: ``list`` of :class:`Zone`
        """
        data = self._paginated_request("/v2/domains", "domains")

        return [self._to_zone(item) for item in data]

    def get_zone(self, zone_id: str) -> Zone:
        """Return a Zone instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :rtype: :class:`Zone`
        """
        resp = self.connection.request("/v2/domains/%s" % zone_id)

        return self._to_zone(resp.object["domain"])

    def create_zone(
        self,
        domain: str,
        type: str = "master",
        ttl: Optional[int] = None,
        extra: Optional[Dict[str, Any]] = None,
    ) -> Zone:
        """Create a new zone.

        :param domain: Zone domain name (e.g. example.com)
        :type domain: ``str``

        :param type: Zone type. Only 'master' value is supported.
        :type  type: ``str``

        :param ttl: TTL for new records. (unused)
        :type  ttl: ``int``

        :param extra: Extra attributes 'ip': ``str`` IP for a default A record
                                       'dns_sec': ``bool`` Enable DSNSEC.
        :type extra: ``dict``

        :rtype: :class:`Zone`
        """

        data = {
            "domain": domain,
        }

        extra = extra or {}

        if "ip" in extra:
            data["ip"] = extra["ip"]

        if "dns_sec" in extra:
            data["dns_sec"] = "enabled" if extra["dns_sec"] is True else "disabled"

        resp = self.connection.request("/v2/domains", data=json.dumps(data), method="POST")

        return self._to_zone(resp.object["domain"])

    def delete_zone(self, zone: Zone) -> bool:
        """Delete a zone.

        Note: This will delete all the records belonging to this zone.

        :param zone: Zone to delete.
        :type  zone: :class:`Zone`

        :rtype: ``bool``
        """
        resp = self.connection.request("/v2/domains/%s" % zone.domain, method="DELETE")

        return resp.success()

    def list_records(self, zone: Zone) -> List[Record]:
        """Return a list of records for the provided zone.

        :param zone: Zone to list records for.
        :type zone: :class:`Zone`

        :return: ``list`` of :class:`Record`
        """
        data = self._paginated_request("/v2/domains/%s/records" % zone.domain, "records")

        return [self._to_record(item, zone) for item in data]

    def get_record(self, zone_id: str, record_id: str) -> Record:
        """Return a Record instance.

        :param zone_id: ID of the required zone
        :type  zone_id: ``str``

        :param record_id: ID of the required record
        :type  record_id: ``str``

        :rtype: :class:`Record`
        """
        resp = self.connection.request("/v2/domains/{}/records/{}".format(zone_id, record_id))

        # Avoid making an extra API call, as zone_id is enough for
        # standard fields
        zone = Zone(id=zone_id, domain=zone_id, type="master", ttl=None, driver=self)

        return self._to_record(resp.object["record"], zone)

    def create_record(
        self,
        name: str,
        zone: Zone,
        type: RecordType,
        data: str,
        extra: Optional[Dict[str, Any]] = None,
    ) -> Record:
        """Create a new record.

        :param name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :param zone: Zone where the requested record is created.
        :type  zone: :class:`Zone`

        :param type: DNS record type (A, AAAA, ...).
        :type  type: :class:`RecordType`

        :param data: Data for the record (depends on the record type).
        :type  data: ``str``

        :keyword extra: Extra attributes 'ttl': Time to live in seconds
                                         'priority': DNS priority. Only
                                                     required for MX and SRV
        :type extra: ``dict``

        :rtype: :class:`Record`
        """
        data = {
            "name": name,
            "type": self.RECORD_TYPE_MAP[type],
            "data": data,
        }
        extra = extra or {}

        if "ttl" in extra:
            data["ttl"] = int(extra["ttl"])

        if "priority" in extra:
            data["priority"] = int(extra["priority"])

        resp = self.connection.request(
            "/v2/domains/%s/records" % zone.domain, data=json.dumps(data), method="POST"
        )

        return self._to_record(resp.object["record"], zone)

    def update_record(
        self,
        record: Record,
        name: Optional[str] = None,
        type: Optional[RecordType] = None,
        data: Optional[str] = None,
        extra: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """Update an existing record.

        :param record: Record to update.
        :type  record: :class:`Record`

        :keyword name: Record name without the domain name (e.g. www).
                     Note: If you want to create a record for a base domain
                     name, you should specify empty string ('') for this
                     argument.
        :type  name: ``str``

        :keyword type: DNS record type. (Unused)
        :type  type: :class:`RecordType`

        :keyword data: Data for the record (depends on the record type).
        :type  data: ``str``

        :keyword extra: Extra attributes 'ttl': Time to live in seconds
                                         'priority': DNS priority. Only
                                                     required for MX and SRV
        :type  extra: ``dict``

        :rtype: ``bool``
        """
        body = {}

        if name:
            body["name"] = name

        if data:
            body["data"] = data

        extra = extra or {}

        if "ttl" in extra:
            body["ttl"] = int(extra["ttl"])

        if "priority" in extra:
            body["priority"] = int(extra["priority"])

        resp = self.connection.request(
            "/v2/domains/{}/records/{}".format(record.zone.domain, record.id),
            data=json.dumps(body),
            method="PATCH",
        )

        return resp.success()

    def delete_record(self, record: Record) -> bool:
        """Delete a record.

        :param record: Record to delete.
        :type  record: :class:`Record`

        :rtype: ``bool``
        """
        resp = self.connection.request(
            "/v2/domains/{}/records/{}".format(record.zone.domain, record.id),
            method="DELETE",
        )

        return resp.success()

    def _to_zone(self, data: Dict[str, Any]) -> Zone:
        type_ = "master"
        domain = data["domain"]
        extra = {
            "date_created": data["date_created"],
        }

        return Zone(id=domain, domain=domain, driver=self, type=type_, ttl=None, extra=extra)

    def _to_record(self, data: Dict[str, Any], zone: Zone) -> Record:
        id_ = data["id"]
        name = data["name"]
        type_ = self._string_to_record_type(data["type"])
        data_ = data["data"]
        ttl = data["ttl"]
        extra = {
            "priority": data["priority"],
        }

        return Record(
            id=id_,
            name=name,
            type=type_,
            data=data_,
            ttl=ttl,
            driver=self,
            zone=zone,
            extra=extra,
        )

    def _paginated_request(
        self,
        url: str,
        key: str,
        params: Optional[Dict[str, Any]] = None,
    ) -> List[Any]:
        """Perform multiple calls to get the full list of items when
        the API responses are paginated.

        :param url: API endpoint
        :type url: ``str``

        :param key: Result object key
        :type key: ``str``

        :param params: Request parameters
        :type params: ``dict``

        :return: ``list`` of API response objects
        :rtype: ``list``
        """
        params = params if params is not None else {}
        resp = self.connection.request(url, params=params).object
        data = list(resp.get(key, []))
        objects = data

        while True:
            next_page = resp["meta"]["links"]["next"]

            if next_page:
                params["cursor"] = next_page
                resp = self.connection.request(url, params=params).object
                data = list(resp.get(key, []))
                objects.extend(data)
            else:
                return objects



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ["Provider", "ContainerState"]


class Type:
    @classmethod
    def tostring(cls, value):
        """Return the string representation of the state object attribute
        :param str value: the state object to turn into string
        :return: the uppercase string that represents the state object
        :rtype: str
        """
        return value.upper()

    @classmethod
    def fromstring(cls, value):
        """Return the state object attribute that matches the string
        :param str value: the string to look up
        :return: the state object attribute that matches the string
        :rtype: str
        """
        return getattr(cls, value.upper(), None)


class Provider:
    """
    Defines for each of the supported providers

    Non-Dummy drivers are sorted in alphabetical order. Please preserve this
    ordering when adding new drivers.
    """

    DUMMY = "dummy"
    DOCKER = "docker"
    ECS = "ecs"
    GKE = "GKE"
    KUBERNETES = "kubernetes"
    LXD = "lxd"
    RANCHER = "rancher"


class ContainerState(Type):
    """
    Standard states for a container

    :cvar RUNNING: Container is running.
    :cvar REBOOTING: Container is rebooting.
    :cvar TERMINATED: Container is terminated.
                This container can't be started later on.
    :cvar STOPPED: Container is stopped.
                This container can be started later on.
    :cvar PENDING: Container is pending.
    :cvar SUSPENDED: Container is suspended.
    :cvar ERROR: Container is an error state.
                Usually no operations can be performed
                on the container once it ends up in the error state.
    :cvar PAUSED: Container is paused.
    :cvar UNKNOWN: Container state is unknown.
    """

    RUNNING = "running"
    REBOOTING = "rebooting"
    TERMINATED = "terminated"
    PENDING = "pending"
    UNKNOWN = "unknown"
    STOPPED = "stopped"
    SUSPENDED = "suspended"
    ERROR = "error"
    PAUSED = "paused"



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Rackspace driver
"""

from libcloud.compute.base import NodeLocation, VolumeSnapshot
from libcloud.compute.types import Provider, LibcloudError, VolumeSnapshotState
from libcloud.utils.iso8601 import parse_date
from libcloud.common.rackspace import AUTH_URL
from libcloud.compute.drivers.openstack import (
    OpenStack_1_0_Response,
    OpenStack_1_0_Connection,
    OpenStack_1_0_NodeDriver,
    OpenStack_1_1_Connection,
    OpenStack_1_1_NodeDriver,
)

SERVICE_TYPE = "compute"
SERVICE_NAME_GEN1 = "cloudServers"
SERVICE_NAME_GEN2 = "cloudServersOpenStack"
ENDPOINT_ARGS_MAP = {
    "dfw": {"service_type": SERVICE_TYPE, "name": SERVICE_NAME_GEN2, "region": "DFW"},
    "ord": {"service_type": SERVICE_TYPE, "name": SERVICE_NAME_GEN2, "region": "ORD"},
    "iad": {"service_type": SERVICE_TYPE, "name": SERVICE_NAME_GEN2, "region": "IAD"},
    "lon": {"service_type": SERVICE_TYPE, "name": SERVICE_NAME_GEN2, "region": "LON"},
    "syd": {"service_type": SERVICE_TYPE, "name": SERVICE_NAME_GEN2, "region": "SYD"},
    "hkg": {"service_type": SERVICE_TYPE, "name": SERVICE_NAME_GEN2, "region": "HKG"},
}


class RackspaceFirstGenConnection(OpenStack_1_0_Connection):
    """
    Connection class for the Rackspace first-gen driver.
    """

    responseCls = OpenStack_1_0_Response
    XML_NAMESPACE = "http://docs.rackspacecloud.com/servers/api/v1.0"
    auth_url = AUTH_URL
    _auth_version = "2.0"
    cache_busting = True

    def __init__(self, *args, **kwargs):
        self.region = kwargs.pop("region", None)
        super().__init__(*args, **kwargs)

    def get_endpoint(self):
        if "2.0" in self._auth_version:
            ep = self.service_catalog.get_endpoint(
                service_type=SERVICE_TYPE, name=SERVICE_NAME_GEN1
            )
        else:
            raise LibcloudError('Auth version "%s" not supported' % (self._auth_version))

        public_url = ep.url

        if not public_url:
            raise LibcloudError("Could not find specified endpoint")

        # This is a nasty hack, but it's required because of how the
        # auth system works.
        # Old US accounts can access UK API endpoint, but they don't
        # have this endpoint in the service catalog. Same goes for the
        # old UK accounts and US endpoint.

        if self.region == "us":
            # Old UK account, which only have uk endpoint in the catalog
            public_url = public_url.replace("https://lon.servers.api", "https://servers.api")
        elif self.region == "uk":
            # Old US account, which only has us endpoints in the catalog
            public_url = public_url.replace("https://servers.api", "https://lon.servers.api")

        return public_url

    def get_service_name(self):
        return SERVICE_NAME_GEN1


class RackspaceFirstGenNodeDriver(OpenStack_1_0_NodeDriver):
    name = "Rackspace Cloud (First Gen)"
    website = "http://www.rackspace.com"
    connectionCls = RackspaceFirstGenConnection
    type = Provider.RACKSPACE_FIRST_GEN
    api_name = "rackspace"

    def __init__(self, key, secret=None, secure=True, host=None, port=None, region="us", **kwargs):
        """
        @inherits:  :class:`NodeDriver.__init__`

        :param region: Region ID which should be used
        :type region: ``str``
        """

        if region not in ["us", "uk"]:
            raise ValueError("Invalid region: %s" % (region))

        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            region=region,
            **kwargs,
        )

    def list_locations(self):
        """
        Lists available locations

        Locations cannot be set or retrieved via the API, but currently
        there are two locations, DFW and ORD.

        @inherits: :class:`OpenStack_1_0_NodeDriver.list_locations`
        """

        if self.region == "us":
            locations = [NodeLocation(0, "Rackspace DFW1/ORD1", "US", self)]
        elif self.region == "uk":
            locations = [NodeLocation(0, "Rackspace UK London", "UK", self)]
        else:
            locations = []

        return locations

    def _ex_connection_class_kwargs(self):
        kwargs = self.openstack_connection_kwargs()
        kwargs["region"] = self.region

        return kwargs


class RackspaceConnection(OpenStack_1_1_Connection):
    """
    Connection class for the Rackspace next-gen OpenStack base driver.
    """

    auth_url = AUTH_URL
    _auth_version = "2.0"

    def __init__(self, *args, **kwargs):
        self.region = kwargs.pop("region", None)
        self.get_endpoint_args = kwargs.pop("get_endpoint_args", None)
        super().__init__(*args, **kwargs)

    def get_service_name(self):
        if not self.get_endpoint_args:
            # if they used ex_force_base_url, assume the Rackspace default

            return SERVICE_NAME_GEN2

        return self.get_endpoint_args.get("name", SERVICE_NAME_GEN2)

    def get_endpoint(self):
        if not self.get_endpoint_args:
            raise LibcloudError("RackspaceConnection must have get_endpoint_args set")

        if "2.0" in self._auth_version:
            ep = self.service_catalog.get_endpoint(**self.get_endpoint_args)
        else:
            raise LibcloudError('Auth version "%s" not supported' % (self._auth_version))

        public_url = ep.url

        if not public_url:
            raise LibcloudError("Could not find specified endpoint")

        return public_url


class RackspaceNodeDriver(OpenStack_1_1_NodeDriver):
    name = "Rackspace Cloud (Next Gen)"
    website = "http://www.rackspace.com"
    connectionCls = RackspaceConnection
    type = Provider.RACKSPACE

    _networks_url_prefix = "/os-networksv2"

    def __init__(
        self,
        key,
        secret=None,
        secure=True,
        host=None,
        port=None,
        region="dfw",
        **kwargs,
    ):
        """
        @inherits:  :class:`NodeDriver.__init__`

        :param region: ID of the region which should be used.
        :type region: ``str``
        """
        valid_regions = ENDPOINT_ARGS_MAP.keys()

        if region not in valid_regions:
            raise ValueError("Invalid region: %s" % (region))

        if region == "lon":
            self.api_name = "rackspacenovalon"
        elif region == "syd":
            self.api_name = "rackspacenovasyd"
        else:
            self.api_name = "rackspacenovaus"

        super().__init__(
            key=key,
            secret=secret,
            secure=secure,
            host=host,
            port=port,
            region=region,
            **kwargs,
        )

    def _to_snapshot(self, api_node):
        if "snapshot" in api_node:
            api_node = api_node["snapshot"]

        extra = {
            "volume_id": api_node["volumeId"],
            "name": api_node["displayName"],
            "created": api_node["createdAt"],
            "description": api_node["displayDescription"],
            "status": api_node["status"],
        }

        state = self.SNAPSHOT_STATE_MAP.get(api_node["status"], VolumeSnapshotState.UNKNOWN)

        try:
            created_td = parse_date(api_node["createdAt"])
        except ValueError:
            created_td = None

        snapshot = VolumeSnapshot(
            id=api_node["id"],
            driver=self,
            size=api_node["size"],
            extra=extra,
            created=created_td,
            state=state,
            name=api_node["displayName"],
        )

        return snapshot

    def _ex_connection_class_kwargs(self):
        endpoint_args = ENDPOINT_ARGS_MAP[self.region]
        kwargs = self.openstack_connection_kwargs()
        kwargs["region"] = self.region
        kwargs["get_endpoint_args"] = endpoint_args

        return kwargs
